#! /usr/bin/env python
# coding=utf-8

import tensorflow as tf

# import tensorflow_addons as tfa
class BatchNormalization(tf.keras.layers.BatchNormalization):
    """
    "Frozen state" and "inference mode" are two separate concepts.
    `layer.trainable = False` is to freeze the layer, so the layer will use
    stored moving `var` and `mean` in the "inference mode", and both `gama`
    and `beta` will not be updated !
    """

    def call(self, x, training=False):
        if not training:
            training = tf.constant(False)
        training = tf.logical_and(training, self.trainable)
        return super().call(x, training)


def convolutional(
    input_layer,
    filters_shape,
    downsample=False,
    activate=True,
    bn=True,
    activate_type="leaky",
):
    if downsample:
        input_layer = tf.keras.layers.ZeroPadding2D(((1, 0), (1, 0)))(input_layer)
        padding = "valid"
        strides = 2
    else:
        strides = 1
        padding = "same"

    conv = tf.keras.layers.Conv2D(
        filters=filters_shape[-1],
        kernel_size=filters_shape[0],
        strides=strides,
        padding=padding,
        use_bias=not bn,
        kernel_regularizer=tf.keras.regularizers.l2(0.0005),
        kernel_initializer=tf.random_normal_initializer(stddev=0.01),
        bias_initializer=tf.constant_initializer(0.0),
    )(input_layer)

    if bn:
        conv = BatchNormalization()(conv)
    if activate == True:
        if activate_type == "leaky":
            conv = tf.nn.leaky_relu(conv, alpha=0.1)
        elif activate_type == "mish":
            conv = mish(conv)
    return conv


def mish(x):
    return x * tf.math.tanh(tf.math.softplus(x))
    # return tf.keras.layers.Lambda(lambda x: x*tf.tanh(tf.math.log(1+tf.exp(x))))(x)


def residual_block(
    input_layer, input_channel, filter_num1, filter_num2, activate_type="leaky"
):
    short_cut = input_layer
    conv = convolutional(
        input_layer,
        filters_shape=(1, 1, input_channel, filter_num1),
        activate_type=activate_type,
    )
    conv = convolutional(
        conv,
        filters_shape=(3, 3, filter_num1, filter_num2),
        activate_type=activate_type,
    )

    residual_output = short_cut + conv
    return residual_output


# def block_tiny(input_layer, input_channel, filter_num1, activate_type='leaky'):
#     conv = convolutional(input_layer, filters_shape=(3, 3, input_channel, filter_num1), activate_type=activate_type)
#     short_cut = input_layer
#     conv = convolutional(conv, filters_shape=(3, 3, input_channel, filter_num1), activate_type=activate_type)
#
#     input_data = tf.concat([conv, short_cut], axis=-1)
#     return residual_output


def route_group(input_layer, groups, group_id):
    convs = tf.split(input_layer, num_or_size_splits=groups, axis=-1)
    return convs[group_id]


def upsample(input_layer):
    return tf.image.resize(
        input_layer,
        (input_layer.shape[1] * 2, input_layer.shape[2] * 2),
        method="bilinear",
    )



#! /usr/bin/env python
# coding=utf-8
from easydict import EasyDict as edict


__C = edict()
# Consumers can get config by: from config import cfg

cfg = __C

# YOLO options
__C.YOLO = edict()

__C.YOLO.CLASSES = "/home/jinwon/Documents/github/factory/packages/aidoop-r/object-tracker-python/examples/yolo-cabbage/custom-cabbage/obj.names"
__C.YOLO.ANCHORS = [
    12,
    16,
    19,
    36,
    40,
    28,
    36,
    75,
    76,
    55,
    72,
    146,
    142,
    110,
    192,
    243,
    459,
    401,
]
__C.YOLO.ANCHORS_V3 = [
    10,
    13,
    16,
    30,
    33,
    23,
    30,
    61,
    62,
    45,
    59,
    119,
    116,
    90,
    156,
    198,
    373,
    326,
]
__C.YOLO.ANCHORS_TINY = [23, 27, 37, 58, 81, 82, 81, 82, 135, 169, 344, 319]
__C.YOLO.STRIDES = [8, 16, 32]
__C.YOLO.STRIDES_TINY = [16, 32]
__C.YOLO.XYSCALE = [1.2, 1.1, 1.05]
__C.YOLO.XYSCALE_TINY = [1.05, 1.05]
__C.YOLO.ANCHOR_PER_SCALE = 3
__C.YOLO.IOU_LOSS_THRESH = 0.5


# Train options
__C.TRAIN = edict()

__C.TRAIN.ANNOT_PATH = "/home/jinwon/Documents/github/factory/packages/aidoop-r/object-tracker-python/examples/yolo-cabbage/custom-cabbage/train.txt"
__C.TRAIN.BATCH_SIZE = 2
# __C.TRAIN.INPUT_SIZE            = [320, 352, 384, 416, 448, 480, 512, 544, 576, 608]
__C.TRAIN.INPUT_SIZE = 416
__C.TRAIN.DATA_AUG = True
__C.TRAIN.LR_INIT = 1e-3
__C.TRAIN.LR_END = 1e-6
__C.TRAIN.WARMUP_EPOCHS = 2
__C.TRAIN.FISRT_STAGE_EPOCHS = 20
__C.TRAIN.SECOND_STAGE_EPOCHS = 30


# TEST options
__C.TEST = edict()

__C.TEST.ANNOT_PATH = "/home/jinwon/Documents/github/factory/packages/aidoop-r/object-tracker-python/examples/yolo-cabbage/custom-cabbage/test.txt"
__C.TEST.BATCH_SIZE = 2
__C.TEST.INPUT_SIZE = 416
__C.TEST.DATA_AUG = False
__C.TEST.DECTECTED_IMAGE_PATH = "/home/jinwon/Documents/github/factory/packages/aidoop-r/object-tracker-python/examples/yolo-cabbage/custom-cabbage/detection/"
__C.TEST.SCORE_THRESHOLD = 0.25
__C.TEST.IOU_THRESHOLD = 0.5



from pyaidoop_dl_models.mask_rcnn import model as modellib, utils
from pyaidoop_dl_models.mask_rcnn.config import Config
import os
import sys
import json
import datetime
import numpy as np
import skimage.draw
import time
import cv2
import cv2.aruco as aruco

############################################################
#  Inference Configurations
############################################################


class InferenceConfig(Config):
    """Configuration for training on the toy  dataset.
    Derives from the base Config class and overrides some values.
    """

    # Give the configuration a recognizable name
    NAME = "object-train"

    # We use a GPU with 12GB memory, which can fit two images.
    # Adjust down if you use a smaller GPU.
    IMAGES_PER_GPU = 1

    # Number of classes (including background)
    NUM_CLASSES = 1 + 1  # Background + balloon

    # Skip detections with < 70% confidence
    DETECTION_MIN_CONFIDENCE = 0.7

    # Backbone network architecture
    # Supported values are: resnet50, resnet101.
    # You can also provide a callable that should have the signature
    # of model.resnet_graph. If you do so, you need to supply a callable
    # to COMPUTE_BACKBONE_SHAPE as well
    BACKBONE = "resnet101"  # "resnet101" or "resnet50"


class MaskRcnnDetect:

    PRINT_DETECTION_TIME = False

    def __init__(self, weight_file_path, log_dir):
        assert (weight_file_path is not None) and (
            log_dir is not None
        ), "weight_file_path and log_dir should not be empty"
        self.weight_file_path = weight_file_path
        self.log_dir = log_dir

        # initialize mask rcnn
        self.inference_config = InferenceConfig()
        self.inference_config.display()

        # create a model
        self.model = modellib.MaskRCNN(
            mode="inference", config=self.inference_config, model_dir=self.log_dir
        )
        assert self.model is not None

        self.model.load_weights(self.weight_file_path, by_name=True)

        # mask list
        self.mask_list = list()

        # scores
        self.scores = list()

    def detect_object_by_path(self, image_path):
        assert image_path is not None

        # Read image
        image = skimage.io.imread(image_path)

        # Detect objects
        sttime = time.time()
        result = self.model.detect([image], verbose=0)[0]
        edtime = time.time()

        if MaskRcnnDetect.PRINT_DETECTION_TIME:
            print("-----------------------------------------------------")
            print("MRCNN Elapsed Time: {} sec".format(edtime - sttime))
            print("Object Count: ", result["masks"].shape[-1])

        self.mask_list = self.get_mask_list(result["masks"])
        self.scores = result["scores"]

        return self.mask_list

        # # Color splash
        # splash = get_mask_by_index(image, result['masks'], 0)

        # # Save output
        # file_name = "splash_{:%Y%m%dT%H%M%S}.png".format(
        #     datetime.datetime.now())
        # skimage.io.imsave(file_name, splash)
        # print("Saved to ", file_name)

        # return splash

    def detect_object_by_data(self, image_data):
        assert image_data is not None, "image_data should not be empty"

        # Detect objects
        sttime = time.time()
        result = self.model.detect([image_data], verbose=0)[0]
        edtime = time.time()

        if MaskRcnnDetect.PRINT_DETECTION_TIME:
            print("-----------------------------------------------------")
            print("MRCNN Elapsed Time: {} sec".format(edtime - sttime))
            print("Object Count: ", result["masks"].shape[-1])

        self.mask_list = self.get_mask_list(result["masks"])
        self.scores = result["scores"]

        return self.mask_list

    def get_mask_list(self, masks):

        # mask_cnt = masks.shape[-1]
        # for idx in range(mask_cnt):
        #     mask = masks[:, :, 0]
        #     mask.reshape(mask.shape[0], mask.shape[1], 1)
        #     mask_gray = np.where(mask, 255, 0).astype(np.uint8)

        mask_list = list()

        # get mask count
        mask_cnt = masks.shape[-1]

        # make the list of mask
        for mask_idx in range(mask_cnt):
            mask = masks[:, :, mask_idx]
            mask_list.append(mask)

        return mask_list

    def get_center_points(self, mask_list):
        center_point_list = list()

        if len(mask_list) <= 0:
            return center_point_list

        if len(mask_list) > 0:
            accumulated_mask = mask_list[0]
            for mask in mask_list:
                accumulated_mask = np.logical_or(mask, accumulated_mask)

            mask_image = np.where(accumulated_mask, 255, 0).astype(np.uint8)

            # get the center point of mask regions
            contours, hierarchy = cv2.findContours(
                mask_image, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE
            )
            for contour in contours:
                # calculate moments for each contour
                M = cv2.moments(contour)

                if M["m00"] == 0.0:
                    continue

                # calculate x,y coordinate of center
                cX = int(M["m10"] / M["m00"])
                cY = int(M["m01"] / M["m00"])

                center_point_list.append((cX, cY))

        return center_point_list

    def get_mask_image(self, mask_list, width, height):
        # center_point_list = list()
        if len(mask_list) > 0:
            accumulated_mask = mask_list[0]
            for mask in mask_list:
                accumulated_mask = np.logical_or(mask, accumulated_mask)

            mask_image = np.where(accumulated_mask, 255, 0).astype(np.uint8)

            # # get the center point of mask regions
            # contours, hierarchy = cv2.findContours(
            #     mask_image, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
            # for contour in contours:
            #     # calculate moments for each contour
            #     M = cv2.moments(contour)

            #     if M["m00"] == 0.0:
            #         continue

            #     # calculate x,y coordinate of center
            #     cX = int(M["m10"] / M["m00"])
            #     cY = int(M["m01"] / M["m00"])

            #     center_point_list.append((cX, cY))

            # skip to draw the ceter point
            # cv2.circle(mask_image, (cX, cY), 5, (0, 0, 0), -1)
        else:
            mask_image = np.zeros((height, width))

        return mask_image.astype(np.uint8)

    def get_splash_image_by_index(self, image, masks, mask_index):

        # Copy color pixels from the original color image where mask is set
        if masks.shape[-1] > 0:
            # We're treating all instances as one, so collapse the mask into one layer
            # mask = (np.sum(mask, -1, keepdims=True) >= 1)
            mask = masks[:, :, mask_index]
            mask.reshape(mask.shape[0], mask.shape[1], 1)
            splash = np.where(mask, 255, 0).astype(np.uint8)
        else:
            splash = gray.astype(np.uint8)
        return splash

    def get_scores(self):
        return self.scores


############################################################
#  Inference Example
############################################################
if __name__ == "__main__":
    import argparse

    # Root directory of the project
    ROOT_DIR = os.path.abspath("./")

    # Import Mask RCNN
    sys.path.append(ROOT_DIR)  # To find local version of the library

    # Path to trained weights file
    COCO_WEIGHTS_PATH = "/home/jinwon/Documents/github/object-tracker-python/logs/object-train20201006T1521/mask_rcnn_object-train_0047.h5"

    # Directory to save logs and model checkpoints, if not provided
    # through the command line argument --logs
    DEFAULT_LOGS_DIR = os.path.join(ROOT_DIR, "logs")

    EXAMPLE_IMAGE_PATH = os.path.join(
        ROOT_DIR, "dataset", "balloon", "val", "14898532020_ba6199dd22_k.jpg"
    )

    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description="infer Mask R-CNN to detect an object."
    )
    parser.add_argument(
        "--weights",
        required=True,
        metavar="/path/to/weights.h5",
        help="Path to weights .h5 file or 'coco'",
    )
    parser.add_argument(
        "--image",
        required=True,
        metavar="path or URL to image",
        help="Image to apply the color splash effect on",
    )
    args = parser.parse_args()

    # print("Weights: ", args.weights)
    # print("Dataset: ", args.dataset)
    # print("Logs: ", args.logs)

    mrdetect = MaskRcnnDetect(COCO_WEIGHTS_PATH, DEFAULT_LOGS_DIR)

    mrdetect.detect_object_by_path(EXAMPLE_IMAGE_PATH)



from setuptools import setup, find_packages

long_description = """A python library for AIdoop web-based application platform such as AIdoop-R and AIdoop-P.
Supports camera operation, object trakcing using tensorflow-based deep learning models, opecv-based image processing 
and graphql and websocket to cooperate with web application server"""

INSTALL_REQUIRES = [
    "numpy",
    "opencv-python",
    "websocket-client",
    "scipy==1.4.1",
    "Pillow",
    "cython",
    "matplotlib",
    "tensorflow==2.5.0",
    "scikit-image==0.16.2",
    "h5py",
    "imgaug",
    "IPython[all]",
]

setup(
    name="pyaidoop-dl-models",
    version="0.1.0",
    author="Jinwon Choi",
    author_email="jinwon@ai-doop.com",
    url="https://github.com/aidoop/pyaidoop.git",
    install_requires=INSTALL_REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "": ["LICENSE", "README.md", "requirements.txt"],
    },
    license="MIT",
    description="AIdoop web-based platfrom library for robot application platform, AIdoop-R",
    long_description=long_description,
    keywords=[
        "cooperative robot",
        "web-based application",
        "object detection",
        "neural network",
        "CNN",
        "machine learning",
        "computer vision",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import sys
import tensorflow as tf
import time

physical_devices = tf.config.experimental.list_physical_devices("GPU")
if len(physical_devices) > 0:
    tf.config.experimental.set_memory_growth(physical_devices[0], True)
from absl import app, flags, logging
from absl.flags import FLAGS
import pyaidoop_dl_models.yolo.utils as utils
from pyaidoop_dl_models.yolo.yolov4 import filter_boxes
from tensorflow.python.saved_model import tag_constants
from PIL import Image
import cv2
import numpy as np
from enum import Enum


from tensorflow.compat.v1 import ConfigProto
from tensorflow.compat.v1 import InteractiveSession

flags.DEFINE_string("framework", "tf", "(tf, tflite, trt")
flags.DEFINE_string("weights", "./checkpoints/yolov4-416", "path to weights file")
flags.DEFINE_integer("size", 416, "resize images to")
flags.DEFINE_boolean("tiny", False, "yolo or yolo-tiny")
flags.DEFINE_string("model", "yolov4", "yolov3 or yolov4")
flags.DEFINE_string("image", "./data/kite.jpg", "path to input image")
flags.DEFINE_string("output", "result.png", "path to output image")
flags.DEFINE_float("iou", 0.45, "iou threshold")
flags.DEFINE_float("score", 0.25, "score threshold")


class YoloDetect:
    FRAMEWORK_TYPE = "tf"
    WEIGHT_PATH = ""
    IMAGE_RESIZE = 416
    TINY = False
    YOLO_TYPE = "yolov4"
    IOU_THRESHOLD = 0.45
    CLASS_SCORE = 0.25

    def __init__(
        self,
        framework_type="tf",
        weight_file_path="",
        image_resize=416,
        yolo_type="yolov4",
    ):
        assert (framework_type is not None) and (
            weight_file_path is not None
        ), "framework type and weight_file_path should not be empty"
        self.framework_type = framework_type
        self.weight_file_path = weight_file_path
        self.image_resize = image_resize
        self.yolo_type = yolo_type

        self.boxes = None
        self.pred_conf = None

        if self.framework_type == "tflite":
            self.model = tf.lite.Interpreter(model_path=FLAGS.weights)
        else:
            self.model = tf.saved_model.load(
                self.weight_file_path, tags=[tag_constants.SERVING]
            )

    def process_detect(self, images_data):
        try:
            if self.framework_type == "tflite":
                self.model = tf.lite.Interpreter(model_path=FLAGS.weights)
                self.model.allocate_tensors()
                input_details = self.model.get_input_details()
                output_details = self.model.get_output_details()
                # print(input_details)
                # print(output_details)
                self.model.set_tensor(input_details[0]["index"], images_data)
                self.model.invoke()
                pred = [
                    self.model.get_tensor(output_details[i]["index"])
                    for i in range(len(output_details))
                ]
                if FLAGS.model == "yolov3" and FLAGS.tiny == True:
                    boxes, pred_conf = filter_boxes(
                        pred[1],
                        pred[0],
                        score_threshold=0.25,
                        input_shape=tf.constant([self.image_resize, self.image_resize]),
                    )
                else:
                    boxes, pred_conf = filter_boxes(
                        pred[0],
                        pred[1],
                        score_threshold=0.25,
                        input_shape=tf.constant([self.image_resize, self.image_resize]),
                    )
            else:
                batch_data = tf.constant(images_data)
                infer = self.model.signatures["serving_default"]

                pred_bbox = infer(batch_data)
                for key, value in pred_bbox.items():
                    boxes = value[:, :, 0:4]
                    pred_conf = value[:, :, 4:]
        except Exception as ex:
            print("Error :", ex, file=sys.stderr)
            boxes = pred_conf = None

        return (boxes, pred_conf)

    def postprocess(self, boxes, pred_conf):
        (
            boxes,
            scores,
            classes,
            valid_detections,
        ) = tf.image.combined_non_max_suppression(
            boxes=tf.reshape(boxes, (tf.shape(boxes)[0], -1, 1, 4)),
            scores=tf.reshape(
                pred_conf, (tf.shape(pred_conf)[0], -1, tf.shape(pred_conf)[-1])
            ),
            max_output_size_per_class=50,
            max_total_size=50,
            iou_threshold=YoloDetect.IOU_THRESHOLD,
            score_threshold=YoloDetect.CLASS_SCORE,
        )
        pred_bbox = [
            boxes.numpy(),
            scores.numpy(),
            classes.numpy(),
            valid_detections.numpy(),
        ]
        return pred_bbox

    def detect_object_by_path(self, image_path):
        image_input_data = cv2.imread(image_path)
        return self.detect_object_by_data(image_input_data)

    def detect_object_by_data(self, image_data):
        image_input_data = cv2.cvtColor(image_data, cv2.COLOR_BGR2RGB)

        image_resize_data = cv2.resize(
            image_input_data, (self.image_resize, self.image_resize)
        )
        image_resize_data = image_resize_data / 255.0

        images_data = []
        for i in range(1):
            images_data.append(image_resize_data)
        images_data = np.asarray(images_data).astype(np.float32)

        # detect objects
        (self.boxes, self.pred_conf) = self.process_detect(images_data)

        pred_bbox = self.postprocess(self.boxes, self.pred_conf)

        return pred_bbox

    def get_box_list(self):
        return self.boxes

    def get_pred_conf_list(self):
        return self.pred_conf
